@@ -1,8 +1,8 @@
 /*!
-# Restricted XML parsing
+# Restricted XML parsing and encoding
 
-This crate provides "restricted" parsing of XML 1.0 documents with
-namespacing.
+This crate provides "restricted" parsing and encoding of XML 1.0 documents
+with namespacing.
 
 ## Features (some call them restrictions)
 
@@ -11,11 +11,13 @@ namespacing.
 * No DTD whatsoever
 * No processing instructions
 * No comments
-* UTF-8 input only
+* UTF-8 only
 * Namespacing-well-formedness enforced
 * XML 1.0 only
 * Streamed parsing (parser emits a subset of SAX events)
-* Can be driven push- and pull-based
+* Streamed encoding
+* Parser can be driven push- and pull-based
+* Tokio-based asynchronicity supported via the `async` feature and [`AsyncParser`].
 
 ## Example
 
@@ -32,7 +34,7 @@ let result = fp.read_all_eof(|ev| {
 assert_eq!(result.unwrap(), true);
 ```
 
-## High-level usage
+## High-level parser usage
 
 ### Push-based usage
 
@@ -44,31 +46,56 @@ in the application and process the resulting [`Event`]s as they happen.
 If the parser should block while waiting for more data to arrive, a
 [`PullParser`] can be used instead. The `PullParser` requires a source which
 implements [`io::BufRead`].
+
+### Usage with Tokio
+
+Tokio is supported with the `async` feature. It offers the [`AsyncParser`]
+and the [`AsyncEventRead`] trait, which work similar to the `PullParser`.
+Instead of blocking, however, the async parser will yield control to other
+tasks.
 */
 #[allow(unused_imports)]
 use std::io;
 
+mod bufq;
+mod context;
+mod errctx;
 pub mod error;
 pub mod lexer;
 pub mod parser;
-mod bufq;
 pub mod strings;
-mod context;
+pub mod writer;
 
 #[cfg(test)]
-mod tests;
+pub mod tests;
 
+#[doc(inline)]
+pub use bufq::BufferQueue;
+pub use context::Context;
 #[doc(inline)]
 pub use error::{Error, Result};
 #[doc(inline)]
-pub use lexer::{Lexer, LexerOptions, DecodingReader};
+pub use lexer::{Lexer, LexerOptions};
 #[doc(inline)]
-pub use parser::{QName, Parser, Event, LexerAdapter, XMLVersion, XMLNS_XML};
+pub use parser::{Event, LexerAdapter, Parser, QName, XMLVersion, XMLNS_XML};
+pub use strings::{CData, CDataStr, NCName, NCNameStr, Name, NameStr};
 #[doc(inline)]
-pub use bufq::BufferQueue;
-pub use strings::{NCName, Name, NCNameStr, NameStr, CData, CDataStr};
-pub use context::Context;
+pub use writer::{Encoder, Item};
 
+#[cfg(feature = "macros")]
+#[doc(inline)]
+#[cfg_attr(docsrs, doc(cfg(feature = "macros")))]
+pub use rxml_proc::{xml_cdata, xml_name, xml_ncname};
+
+#[cfg(feature = "async")]
+mod future;
+
+#[cfg(feature = "async")]
+#[doc(inline)]
+#[cfg_attr(docsrs, doc(cfg(feature = "async")))]
+pub use future::{AsyncEventRead, AsyncEventReadExt, AsyncParser};
+
+/// Package version
 pub const VERSION: &'static str = env!("CARGO_PKG_VERSION");
 
 /**
@@ -96,7 +123,8 @@ pub trait EventRead {
 	/// returned again by the parser on the next invocation without reading
 	/// further data from the source).
 	fn read_all<F>(&mut self, mut cb: F) -> Result<()>
-		where F: FnMut(Event) -> ()
+	where
+		F: FnMut(Event) -> (),
 	{
 		loop {
 			match self.read()? {
@@ -119,7 +147,8 @@ pub trait EventRead {
 	/// returned again by the parser on the next invocation without reading
 	/// further data from the source).
 	fn read_all_eof<F>(&mut self, cb: F) -> Result<bool>
-		where F: FnMut(Event) -> ()
+	where
+		F: FnMut(Event) -> (),
 	{
 		as_eof_flag(self.read_all(cb))
 	}
@@ -147,8 +176,8 @@ fp.feed(doc[..10].to_vec());
 // We expect a WouldBlock, because the XML declaration is not complete yet
 let ev = fp.read();
 assert!(matches!(
-    ev.err().unwrap(),
-    Error::IO(e) if e.kind() == io::ErrorKind::WouldBlock
+	ev.err().unwrap(),
+	Error::IO(e) if e.kind() == io::ErrorKind::WouldBlock
 ));
 
 fp.feed(doc[10..25].to_vec());
@@ -159,7 +188,7 @@ assert!(matches!(ev.unwrap().unwrap(), Event::XMLDeclaration(_, XMLVersion::V1_0
 ```
 */
 pub struct FeedParser<'x> {
-	token_source: LexerAdapter<DecodingReader<BufferQueue<'x>>>,
+	token_source: LexerAdapter<BufferQueue<'x>>,
 	parser: Parser,
 }
 
@@ -185,11 +214,8 @@ impl<'x> FeedParser<'x> {
 	}
 
 	pub fn with_context(ctx: parser::RcPtr<Context>) -> FeedParser<'x> {
-		FeedParser{
-			token_source: LexerAdapter::new(
-				Lexer::new(),
-				DecodingReader::new(BufferQueue::new()),
-			),
+		FeedParser {
+			token_source: LexerAdapter::new(Lexer::new(), BufferQueue::new()),
 			parser: Parser::with_context(ctx),
 		}
 	}
@@ -205,9 +231,8 @@ impl<'x> FeedParser<'x> {
 	/// # Panics
 	///
 	/// If [`FeedParser::feed_eof()`] has been called before.
-	pub fn feed<'a: 'x, T: Into<std::borrow::Cow<'a, [u8]>>>(&mut self, data: T)
-	{
-		self.token_source.get_mut().get_mut().push(data);
+	pub fn feed<'a: 'x, T: Into<std::borrow::Cow<'a, [u8]>>>(&mut self, data: T) {
+		self.token_source.get_mut().push(data);
 	}
 
 	/// Feed the eof marker to the parser.
@@ -219,7 +244,7 @@ impl<'x> FeedParser<'x> {
 	/// After the eof marker has been fed to the parser, no further data can
 	/// be fed.
 	pub fn feed_eof(&mut self) {
-		self.token_source.get_mut().get_mut().push_eof();
+		self.token_source.get_mut().push_eof();
 	}
 
 	/// Return the amount of bytes which have not been read from the buffer
@@ -230,7 +255,7 @@ impl<'x> FeedParser<'x> {
 	/// to `feed()`) has been processed (and only if that chunk is owned by
 	/// the parser).
 	pub fn buffered(&self) -> usize {
-		self.token_source.get_ref().get_ref().len()
+		self.token_source.get_ref().len()
 	}
 
 	/// Return a reference to the internal buffer BufferQueue
@@ -238,7 +263,7 @@ impl<'x> FeedParser<'x> {
 	/// This can be used to force dropping of all memory in case of error
 	/// conditions.
 	pub fn get_buffer_mut(&mut self) -> &mut BufferQueue<'x> {
-		self.token_source.get_mut().get_mut()
+		self.token_source.get_mut()
 	}
 
 	/// Release all temporary buffers
@@ -278,6 +303,16 @@ data available (or returns an error).
 Interaction with a `PullParser` should happen exclusively via the
 [`EventRead`] trait.
 
+## Blocking I/O
+
+If the [`PullParser`] is used with blocking I/O and a source which may block for a significant amount of time (e.g. a network socket), some events may be emitted with significant delay. This is due to an edge case where the lexer may emit a token without consuming a byte from the source.
+
+This internal state of the lexer is not observable from the outside, but it affects most importantly closing element tags. In practice, this means that the last closing element tag of a "stanza" of XML is only going to be emitted once the first byte of the next stanza has been made available through the BufRead.
+
+This only affects blocking I/O, because a non-blocking source will return [`std::io::ErrorKind::WouldBlock`] from the read call and yield control back to the parser to emit the event.
+
+In general, for networked operations, it is recommended to use the [`FeedParser`] or [`AsyncParser`] instead of the [`PullParser`].
+
 ## Example
 
 ```
@@ -292,29 +327,33 @@ let ev = pp.read();
 assert!(matches!(ev.unwrap().unwrap(), Event::XMLDeclaration(_, XMLVersion::V1_0)));
 ```
 */
-pub struct PullParser<T: io::Read + Sized> {
-	token_source: LexerAdapter<DecodingReader<T>>,
+pub struct PullParser<T: io::BufRead> {
 	parser: Parser,
+	token_source: LexerAdapter<T>,
 }
 
-impl<T: io::Read + Sized> PullParser<T> {
-	/// Create a new PullParser, wrapping the given reader.
-	///
-	/// **Note:** It is highly recommended to wrap a common reader into
-	/// [`std::io::BufReader`] as the implementation will do lots of small
-	/// `read()` calls. Those would be terribly inefficient without buffering.
-	pub fn new(r: T) -> Self {
-		PullParser{
-			token_source: LexerAdapter::new(
-				Lexer::new(),
-				DecodingReader::new(r),
-			),
-			parser: Parser::new(),
+impl<T: io::BufRead> PullParser<T> {
+	/// Create a new parser with default options, wrapping the given reader.
+	pub fn new(inner: T) -> Self {
+		Self::with_options(inner, LexerOptions::default())
+	}
+
+	/// Create a new parser while configuring the lexer with the given
+	/// options.
+	pub fn with_options(inner: T, options: LexerOptions) -> Self {
+		Self::wrap(inner, Lexer::with_options(options), Parser::new())
+	}
+
+	/// Create a fully customized parser from a lexer and a parser component.
+	pub fn wrap(inner: T, lexer: Lexer, parser: Parser) -> Self {
+		Self {
+			token_source: LexerAdapter::new(lexer, inner),
+			parser,
 		}
 	}
 }
 
-impl<T: io::Read + Sized> EventRead for PullParser<T> {
+impl<T: io::BufRead> EventRead for PullParser<T> {
 	/// Read a single event from the parser.
 	///
 	/// If the EOF has been reached with a valid document, `None` is returned.
